/*************************************************************************
 * Copyright (C) 1995-2000, Rene Brun and Fons Rademakers.               *
 * All rights reserved.                                                  *
 *                                                                       *
 * For the licensing terms see $ROOTSYS/LICENSE.                         *
 * For the list of contributors see $ROOTSYS/README/CREDITS.             *
 *************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <fstream>
#include <iostream>

#include "TROOT.h"
#include "TClass.h"
#include "TLTFPainter.h"
#include "TH2.h"
#include "TF2.h"
#include "TF3.h"
#include "TVirtualPad.h"
#include "TPaveStats.h"
#include "TFrame.h"
#include "TText.h"
#include "TLine.h"
#include "TPolyLine.h"
#include "TPoints.h"
#include "TStyle.h"
#include "TGraph.h"
#include "TGaxis.h"
#include "TColor.h"
#include "TLego.h"
#include "TView.h"
#include "TMath.h"
#include "TRandom.h"
#include "TObjArray.h"
#include "Hparam.h"

#include "TApplication.h"
#include "TCanvas.h"
#include "TTimer.h"
#include "TGClient.h"
#include "TGX11.h"

using namespace std;

#include "RVersion.h"
#if ROOT_VERSION_CODE >= ROOT_VERSION(5,0,0)
  #define ROOT_V5 1
#endif

//#define _NO_DOUBLEBUF

   extern Hparam_t  Hparam;

   const Int_t kScrollEvent = 9988;

   //ClassImp(TLTFPainter)
   //ClassImp(TLTimeFrequency)



   // A kludge!
   class MyX11 : public TGX11 {
   public:
      MyX11 () {
      }
   
      void CopyArea (Int_t src_x, Int_t src_y, UInt_t width, UInt_t height,
                    Int_t dest_x, Int_t dest_y) {      
#ifndef ROOT_V5
         XWindow_t* w = GetCurrentWindow();
         TGX11::CopyArea (w->drawing, w->drawing, (GContext_t)*GetGC (2), 
#else /* ROOT_V5 */
	//XWindow_t* w = GetCurrentWindow();
	 Window_t w = GetCurrentWindow();
         TGX11::CopyArea (w, w, (GContext_t)*GetGC (2), 
#endif /* ROOT_V5 */
                         src_x, src_y, width, height, dest_x, dest_y); }
      void ClearArea (Int_t x1, Int_t y1, Int_t ww, Int_t hh) {
      #ifdef _NO_DOUBLEBUF
#ifndef ROOT_V5
         XWindow_t* w = GetCurrentWindow();
         TGX11::ClearArea (w->window, x1, y1, ww, hh); 
#else /* ROOT_V5 */
	//XWindow_t* w = GetCurrentWindow();
         TGX11::ClearArea (w, x1, y1, ww, hh); 
#endif /* ROOT_V5 */
      #else
         SetFillColor (0); // white
         SetFillStyle (1); // solid
         gVirtualX->DrawBox (x1, y1, x1 + ww, y1 + hh, kFilled);
      #endif
      }
   };

//______________________________________________________________________________
   TLTFPainter::TLTFPainter()
   {
//:KLUDGE:
//cout << "Entering TLTFPainter::TLTFPainter.\n";

      fHScroll = kFALSE;
      fHScrollNum = 0;
   }

//______________________________________________________________________________
   TLTFPainter::~TLTFPainter()
   {
//:KLUDGE:
//cout << "Entering TLTFPainter::~TLTFPainter.\n";

   }

//______________________________________________________________________________
   Int_t TLTFPainter::TableInit(void) {
//:KLUDGE:
//cout << "Entering TLTFPainter::TableInit.\n";
// Code crashes when calling THistPainter::TableInit() below following mouse click.

      if (!fHScroll) {
         cout << "slow update" << endl;
         return THistPainter::TableInit();
//:KLUDGE:
//       if (gPad->IsEditable()) {
//          //cout << "slow update" << endl;
//          return THistPainter::TableInit();
//       } else { 
//          //cout << "Ignored Request.\n";
//          return 1;
//       }
      }
      // Get full range
      Double_t xmin, xmax, ymin, ymax;
      gPad->GetRangeAxis (xmin, ymin, xmax, ymax);
      // cout << "xmin     = " << xmin << endl;
      // cout << "xmax     = " << xmax << endl;
      // cout << "ymin     = " << ymin << endl;
      // cout << "ymax     = " << ymax << endl;
      // set new paint area
      TAxis oldaxis;
      if (fXaxis) oldaxis.ImportAttributes(fXaxis);

      Int_t first, last, newfirst, newlast;
      first           = fXaxis->GetFirst();
      last            = fXaxis->GetLast();
      if (fHScrollNum > 0) {
         // scroll left
         newfirst	 = last - fHScrollNum + 1;
         newlast 	 = last;
      }
      else {
         // scroll right
         newfirst	 = first;
         newlast 	 = first + fHScrollNum - 1;
      }
      // cout << "first    = " << first << endl;
      // cout << "last     = " << last << endl;
      // cout << "newfirst = " << newfirst << endl;
      // cout << "newlast  = " << newlast << endl;
      // check if scroll is required
      Int_t ret = 0;
      if ((newfirst > first) || (newlast < last)) {
         // set new range
         fXaxis->SetRange (newfirst, newlast);
      }
      // init table
      ret = THistPainter::TableInit();
      // now scroll first before repainting new area
      if ((newfirst > first) || (newlast < last)) {
         if (fHScrollNum > 0) {
            // scroll left
            Int_t px0 = gPad->XtoPixel (xmin);
            Int_t px1 = gPad->XtoPixel (Hparam.xmin);
            Int_t px2 = gPad->XtoPixel (xmax);
            Int_t w = px1 - px0;
            Int_t py2 = gPad->YtoPixel (ymin);
            Int_t py1 = gPad->YtoPixel (ymax);
            Int_t h = py2 - py1;
            // cout << "px0      = " << px0 << endl;
            // cout << "px1      = " << px1 << endl;
            // cout << "px2      = " << px2 << endl;
            // cout << "py1      = " << py1 << endl;
            // cout << "py2      = " << py2 << endl;
            ((MyX11*)gVirtualX)->CopyArea (px2 - w, py1, w, h, px0, py1);
            ((MyX11*)gVirtualX)->ClearArea (px1, py1, px2-px1, h);
         }
         else if (fHScrollNum < 0) {
            // scroll right
            Int_t px0 = gPad->XtoPixel (xmin);
            Int_t px1 = gPad->XtoPixel (Hparam.xmax);
            Int_t px2 = gPad->XtoPixel (xmax);
            Int_t w = px2 - px1;
            Int_t py2 = gPad->YtoPixel (ymin);
            Int_t py1 = gPad->YtoPixel (ymax);
            Int_t h = py2 - py1;
            // cout << "px0      = " << px0 << endl;
            // cout << "px1      = " << px1 << endl;
            // cout << "px2      = " << px2 << endl;
            // cout << "py1      = " << py1 << endl;
            // cout << "py2      = " << py2 << endl;
            ((MyX11*)gVirtualX)->CopyArea (px0, py1, w, h, px1, py1);
            ((MyX11*)gVirtualX)->ClearArea (px0, py1, px1-px0, h);
         }
      }
      // restore old settings
      fXaxis->ImportAttributes(&oldaxis);
      fHScroll = kFALSE;
      fHScrollNum = 0;
      // {
         // static int n = 500000;
         // if (--n < 0) exit (1);
      // }
      return ret;
   }

//______________________________________________________________________________
   void TLTFPainter::HScroll (Int_t bins)
   {
//:KLUDGE:
//cout << "Entering TLTFPainter::HScroll.\n";

      if (!bins) {
         return;
      }
      if (fHScroll) {
         fHScrollNum += bins;
      }
      else {
         fHScroll = kTRUE;
         fHScrollNum = bins;
      }
   }

//______________________________________________________________________________
   void TLTFPainter::ExecuteEvent (Int_t event, Int_t px, Int_t py)
   {
//:KLUDGE:
//cout << "Entering TLTFPainter::ExecuteEvent.\n";

      if (event == kScrollEvent) {
         if (px) HScroll (px);
      }
      else {
         THistPainter::ExecuteEvent (event, px, py);
      }
   }


//______________________________________________________________________________
   TLTimeFrequency::TLTimeFrequency (tftype_t tftype,
                     const char* name, const char* title, 
                     Int_t nbinsx, Axis_t xlow, Axis_t xup, 
                     Int_t nbinsy, Axis_t ylow, Axis_t yup) 
   : TH2D(name, title, nbinsx, xlow, xup, nbinsy, ylow, yup),
     fTfType (kTfColor), fDrawCanv (0)
   {
//:KLUDGE:
//cout << "Entering TLTimeFrequency::TLTimeFrequency.\n";

      SetStats (kFALSE);
      SetTfType (tftype);
      fXaxis.SetTickLength (-fXaxis.GetTickLength()/2);
      fYaxis.SetTickLength (-fYaxis.GetTickLength()/2);
      fXaxis.SetLabelOffset (2 * fXaxis.GetLabelOffset());
      fYaxis.SetLabelOffset (3 * fYaxis.GetLabelOffset());
      fXaxis.SetTitleOffset (1.1 * fXaxis.GetTitleOffset());
      fYaxis.SetTitleOffset (1.2 * fYaxis.GetTitleOffset());
   }

//______________________________________________________________________________
   void TLTimeFrequency::SetTfType (tftype_t tftype)
   {
//:KLUDGE:
//cout << "Entering TLTimeFrequency::SetTfType.\n";

      fTfType = tftype;
      switch (tftype) {
         case kTfColor:
         case kTfColorZ:
         default:
            {
               fDrawOpt = "COL";
               break;
            }
         case kTfScatter:
            {
               fDrawOpt = "SCAT";
               break;
            }
         case kTfBox:
            {
               fDrawOpt = "BOX";
               break;
            }
      }
   }

//______________________________________________________________________________
   void TLTimeFrequency::Draw (Option_t* option)
   {
//:KLUDGE:
//cout << "Entering TLTimeFrequency::Draw.\n";

      fDrawCanv = gPad;
      TString opt = fDrawOpt;
      if (fTfType == kTfColorZ) opt += "Z";
      TH2D::Draw ((const char*)opt);
   }

//______________________________________________________________________________
   void TLTimeFrequency::Paint (Option_t* option) 
   {
//:KLUDGE:
//cout << "Entering TLTimeFrequency::Paint.\n";

      if (!fPainter) {
         //fPainter = GetTFPainter (this);
         fPainter = new TLTFPainter();
         fPainter->SetHistogram (this);
      }
      TH2D::Paint (option);
   }

//______________________________________________________________________________
   void TLTimeFrequency::NewF (Double_t* y)
   {
//:KLUDGE:
//cout << "Entering TLTimeFrequency::NewF.\n";

      if (GetDimension() != 2) {
         return;
      }
      Int_t nx = fXaxis.GetNbins() + 2;
      Int_t ny = fYaxis.GetNbins() + 2;
      Double_t* row = GetArray() + nx;
      for (int i = 1; i <= ny - 2; ++i, row += nx) {
         memmove (row + 1, row + 2, (nx - 2) * sizeof (Double_t));
         row[nx - 2] = y[i - 1];
      }
      if (fPainter) fPainter->ExecuteEvent (kScrollEvent, 1, 0);
   }

//______________________________________________________________________________
   void TLTimeFrequency::Update ()
   {
//:KLUDGE:
//cout << "Entering TLTimeFrequency::Update.\n";

      TVirtualPad* canv = GetCanvas();
      if (!canv) {
         return;
      }
      TVirtualPad* padsave = gPad;
      canv->cd();
   #ifdef _NO_DOUBLEBUF
      canv->GetCanvas()->SetDoubleBuffer (0);
   #else
      canv->GetCanvas()->SetDoubleBuffer (1);
   #endif
      gVirtualX->SetDrawMode (TVirtualX::kCopy);
      TString opt = "SAME" + fDrawOpt;
      Paint ((const char*)opt);
   #ifndef _NO_DOUBLEBUF
      canv->GetCanvas()->Flush();
   #endif
      canv->SetCursor (kCross);
      gPad = padsave;
   }

//______________________________________________________________________________
   TVirtualPad* TLTimeFrequency::GetCanvas()
   {
//:KLUDGE:
//cout << "Entering TLTimeFrequency::GetCanvas.\n";

      if (!fDrawCanv) {
         return 0;
      }
      TVirtualPad* c = 
         (TVirtualPad*)gROOT->GetListOfCanvases()->FindObject (fDrawCanv);
      if (!c) fDrawCanv = 0;
      return fDrawCanv;
   }

//______________________________________________________________________________
   static TClass* gTfPainter = 0;

//______________________________________________________________________________
   TVirtualHistPainter* TLTimeFrequency::GetTFPainter (TLTimeFrequency* tf)
   {
//:KLUDGE:
//cout << "Entering TLTimeFrequency::GetTFPainter.\n";

      if (!gTfPainter) {
         if (gROOT->LoadClass ("TLTFPainter", "TFPainter")) {
            return 0;
         }
         gTfPainter = gROOT->GetClass ("TLTFPainter");
         if (!gTfPainter) {
            return 0;
         }
      }
      TVirtualHistPainter* p = (TVirtualHistPainter*)gTfPainter->New();
      if (p) {
         p->SetHistogram (tf);
      }
      return p;
   }

//______________________________________________________________________________
   Int_t kMaxX = 200;
   Int_t kMaxY = 200;

//______________________________________________________________________________
   class NewData : public TObject {
   protected:
      TLTimeFrequency&	fTF;
      Double_t		fX;
      Double_t		fT;
   public:
      NewData (TLTimeFrequency& tf) : fTF (tf), fX (0.), fT (0.) {
      }
      // Calulate new frequency series and add it
      virtual Bool_t HandleTimer (TTimer* timer)
      {
         Double_t y [kMaxY];
         Double_t x = fX;
         for (int i = 0; i < kMaxY; i++, x += 0.042) {
            Float_t tmp = TMath::Sin (x) + 1.;
            y[i] = tmp * tmp / 4.4 + gRandom->Rndm()/10.;
         }
         fX += TMath::Sin (fT)/50. + 0.015;
         fT += 0.02;
         fTF.NewF (y);
         fTF.Update();
         timer->Reset();
         return kTRUE;
      }
   };

//______________________________________________________________________________

/*

   int main (int argc, char** argv)
   {
      TROOT root ("GUI", "TF Test");
      TApplication theApp ("TF Test", &argc, argv);
   
      new TCanvas ();
      TLTimeFrequency* tf = 
         new TLTimeFrequency (kTfColorZ, "TF", "TF Test", 
                           kMaxX, -kMaxX, 0, kMaxY, 0, kMaxY);
      for (int i = 0; i < kMaxX; i++) {
         for (int j = 0; j < kMaxY; j++) {
            tf->Fill (i-kMaxX, j, 0.5);
         }
      }
      tf->SetMinimum (0.);
      tf->SetMaximum (1.);
      tf->SetXTitle ("Time (s)");
      tf->SetYTitle ("Frequency (Hz)");
      gStyle->SetPalette (1, 0);
      tf->Draw ();
      NewData* nd = new NewData (*tf);
   
      TTimer* timer = new TTimer (nd, 1, kTRUE);
      timer->TurnOn();
      timer->Reset();
   
      theApp.Run();
   
      return 0;
   }

*/
